#!/bin/sh
:
: Apple Open Systems Software Installer
:
#	@(#)Revision is $Id: mae_installer,v 2.1.1.1 1994/04/29 01:48:41 brianw Exp $

# Copyright 1993 Apple Computer, Inc.
# All Rights Reserved.
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF APPLE COMPUTER, INC.
# The copyright notice above does not evidence any actual or
# intended publication of such source code.
#

#------------------------------------------------------------------------------
# Save argc and arg1 as globals
#------------------------------------------------------------------------------
argc=$#                                 # Number of command line args
arg1=${1:-''}                           # Arg list - NULL if not set

#------------------------------------------------------------------------------
# Debugging options set here
#------------------------------------------------------------------------------
# set -u
# set -x

#------------------------------------------------------------------------------
# Initialize some global variables 
#------------------------------------------------------------------------------
in_quit=''
top_tag='TOP'                           # Magic cookie file to tag top dir

#--------------------------------------------------------------------------
# Load terminal specific escape sequences into variables
#--------------------------------------------------------------------------
bld=`tput bold 2> /dev/null`            # Turn BOLD atribute ON 
rev=`tput rev 2> /dev/null`             # Turn REVERSE atribute ON 
off=`tput sgr0 2> /dev/null`            # Turn ALL atributes OFF 
clr_scrn=`tput clear 2>/dev/null`       # Clear screen 

#------------------------------------------------------------------------------
# Set umask and PATH (include cwd), and save the default IFS
#------------------------------------------------------------------------------
umask 0022
PATH=/bin:/usr/bin:/usr/ucb:/etc:/usr/etc:/usr/5bin:.
S_IFS="$IFS"                            # Save a copy of a "good" IFS

#------------------------------------------------------------------------------
# Save Location and Name information
#------------------------------------------------------------------------------
start_dir="`/bin/pwd`"                  # Directory where user was at launch
launch_dir="`dirname \"$0\"`"           # Directory where installer lives
if [ "$launch_dir" = '.' ]; then        # Make sure that it is an absolute path
    launch_dir="$start_dir"     
elif [ "`expr \"$launch_dir\"// : '\/.*'`" -eq 0 ]; then
    if [ "$start_dir" = '/' ]; then
        launch_dir="${start_dir}$launch_dir"  
    else
        launch_dir="${start_dir}/$launch_dir"  
    fi
fi
installer="`basename \"$0\"`"           # Name of install script

#------------------------------------------------------------------------------
# Find the correct strings file 
#------------------------------------------------------------------------------
catalog_dir="${LC_ALL:-${LC_MESSAGES:-${LANG:-C}}}"
eval catalog_dir=\"`echo $catalog_dir | tr "[a-z]" "[A-Z]"`\" 

if [ ! -r "${launch_dir}/LOCALE/${catalog_dir}/INSTALL.STR" ]; then
    echo "
    WARNING:
    Cannot locate message catalog for '$catalog_dir'
    All console messages will appear in English" >&2
    echo "
    Press RETURN to continue or CONTROL-C to quit."
    read i_am_waiting                             # Accept any input and continue

    catalog_dir="C"           # Try finding English strings
    if [ ! -r "${launch_dir}/LOCALE/${catalog_dir}/INSTALL.STR" ]; then
        echo "
    ERROR: Cannot read strings in  '${launch_dir}/LOCALE/${catalog_dir}'"
        exit 5
    fi
fi

. "${launch_dir}/LOCALE/${catalog_dir}/INSTALL.STR"  # Load localized strings

#------------------------------------------------------------------------------
# Save tty state 
#------------------------------------------------------------------------------
save_stty="`/bin/stty -g`"                        # Save stty settings
/bin/stty intr "^C"                               # Ensure that we know what sends SIGINT

#------------------------------------------------------------------------------
# Traps 
#------------------------------------------------------------------------------
trap 'CleanUpExit 3' 1 2 3 15                     # Clean up partial installation

#------------------------------------------------------------------------------
# Clear Functions 
#------------------------------------------------------------------------------
unset -f cd rmdir

###############################################################################
## Function definitions begin here                                           ##
###############################################################################

#------------------------------------------------------------------------------
#   Alert - display message
#------------------------------------------------------------------------------
Alert()
{
    if [ $# -ne 1 ]; then
        AssertErr "No args passed to Alert function."
    fi
    _alert_msg="$1"

    if Interactive; then                # Interactive
        TopScreen "$hit_ret_banner_msg"
        echo "

    ${bld}ERROR: $_alert_msg 
" 
        WaitToContinue;
    else                                # Not Interactive
        echo "
    ERROR: $_alert_msg "  >&2
    fi
}

#------------------------------------------------------------------------------
# AskDir - Prompt for a directory name.  Validate before returning.
#------------------------------------------------------------------------------
AskDir()
{
    if [ $# -ne 4 ]; then
        AssertErr "Incorrect number of args passed to AskDir()."
    fi

    _ask_dir_prompt="$1"
    _ask_dir_help="$2"
    _ask_dir_pkg_name="$3"
    _ask_dir_write_req="$4"                       # Either 'W' or NULL
    eval _ask_dir_def_loc="\$d_${_ask_dir_pkg_name}_dir"
	
    while :; do
        TopScreen;                            # Clear screen and display banner
        echo "$_ask_dir_prompt" 
        if [ -z "$_ask_dir_def_loc" ]; then
            Prompt "    [$bld$no_default_msg$off] >>> "
        else
            Prompt "    [$default_equals_msg$bld$_ask_dir_def_loc$off] >>> "
        fi
        read _ask_dir_location
    
        ToLower _ask_dir_location                          # Convert to Lowercase
        case "$_ask_dir_location_lwr" in
            '')  if [ -z "$_ask_dir_def_loc" ]; then
                     Alert "$alert_no_default_msg"
                     continue                     # No default, ask again
                 else 
                     _ask_dir_location="$_ask_dir_def_loc"          # Else use default
                 fi
                 ;;
            "$quit_msg"|\
            "$quit_short_msg")   AskQuit "$yes_msg"; continue ;;
            "$help_msg"|\
            "$help_short_msg")   DoHelp "$_ask_dir_help"; continue ;;
        esac

        # Run validation script
        "${source_dir}/${_ask_dir_pkg_name}.VAL" "$_ask_dir_location" destination 2>/dev/null &       
        wait $!                       # Need to background it to get PID
        _ask_dir_pkg_result=$?
        if [ $_ask_dir_pkg_result -eq 1 ]; then    # Error
            if [ -r /tmp/$!.MSG ]; then
                Interactive && TopScreen "$hit_ret_banner_msg";
                echo
                cat /tmp/$!.MSG           # Display messages
                echo
                WaitToContinue
                rm -f /tmp/$!.MSG         # Clear to prevent concatenation
            else
                Alert "$an_error_msg (destination)"  # Generic msg
            fi
            return 1
        fi                            
                                              # Don't quote $_ask_dir_write_req !! -ped-
        if BadDir "$_ask_dir_location" "$_ask_dir_pkg_name" $_ask_dir_write_req; then
            if [ "$_ask_dir_location" = "$_ask_dir_def_loc" ]; then
                _ask_dir_def_loc=                          # Default didn't work; reset
                eval ${_ask_dir_pkg_name}_dir=
                eval d_${_ask_dir_pkg_name}_dir=
            fi
            return 1                              # Bad dir, return ERROR
        fi
        return 0                                  # Dir ok, return SUCCESS
    done
}

#------------------------------------------------------------------------------
# AskInstMode - Sets inst_mode
#------------------------------------------------------------------------------
AskInstMode()
{
    if [ -z "$d_inst_mode" ]; then
        AssertErr "No default installation mode in AskInstmode"
    fi
	
    while :; do
        TopScreen;
        echo "$instmode_prompt_msg" 
        echo
        Prompt "    [default=$bld$d_inst_mode$off] >>> "
        read _inst_mode_selection
        ToLower _inst_mode_selection;
        case "$_inst_mode_selection_lwr" in
            '')         inst_mode="$d_inst_mode"
                        return 0
                        ;;
            "$easy_msg"|\
            "$easy_short_msg")     inst_mode="$easy_msg"
                                   d_inst_mode="$inst_mode"
                                   return 0
                                   ;;
            "$custom_msg"|\
            "$custom_short_msg")   inst_mode="$custom_msg"
                                   d_inst_mode="$inst_mode"
                                   return 0
                                   ;;
            "$quit_msg"|\
            "$quit_short_msg")     AskQuit "$yes_msg"; continue  ;;
            "$help_msg"|\
            "$help_short_msg")     DoHelp "$instmode_help_msg"; continue ;;
            *)                     Alert "$alert_not_valid_msg"; continue ;; 
        esac
    done
}

#------------------------------------------------------------------------------
# AskInstRem - Ask if they want to install a particular package
#------------------------------------------------------------------------------
AskInstRem()
{
    if [ $# -ne 3 ]; then
        AssertErr "Incorrect number of arguments passed to AskInstRem()"
    fi
    _ask_inst_prompt="$1"                           # Prompt
    _ask_inst_help="$2"                             # Help
    _ask_inst_pkg_name="$3"                         # Var name

    eval _ask_inst_default="\${inst_$_ask_inst_pkg_name:-"$yes_msg"}" 
    
    AskYesNo "$_ask_inst_prompt" "$_ask_inst_help" "$_ask_inst_default"

    case "$reply" in
        "$yes_msg")  eval inst_$_ask_inst_pkg_name="$yes_msg" ;;
        "$no_msg")   eval inst_$_ask_inst_pkg_name="$no_msg" ;;
    esac

    return 0
}

#------------------------------------------------------------------------------
#   AskQuit - either quit or continue with the install based on input
#------------------------------------------------------------------------------
AskQuit()
{
    if [ $# -ne 1 ]; then
        AssertErr "Incorrect arg count in AskQuit()."
    fi
    quit_default="$1"

    if Interactive; then
        in_quit=TRUE;                   # Prevent recursive calls to AskQuit
        AskYesNo "$ask_quit_msg" "$quit_help_msg" "$quit_default"
        in_quit=''                      # Reset flag
        case "$reply" in
            "$yes_msg") CleanUpExit 0 ;; # Clean up and exit
            "$no_msg")  return 0 ;; # Continue with installation
            *) AssertErr "Invalid return from AskYesNo() in AskQuit()" ;;
        esac
    else
         CleanUpExit 0                  # Just exit if NON-interactive
    fi
}   

#------------------------------------------------------------------------------
# FindSourceDir - Sets source_dir
#------------------------------------------------------------------------------
FindSourceDir()
{
    echo "$find_pkgs_msg"               # Let them know what we are doing
    # Do a Pre-check to try to find the source dir
    if [ ${INSTALL_SRC:-NULL} != NULL ]; then     # Check environment variable 
        if [ -f "${INSTALL_SRC}/${o_s}.${top_tag}" ]; then
            d_source_dir="$INSTALL_SRC"
            return 0                              # Happy, Happy, Joy, Joy
        elif [ -f "$INSTALL_SRC/${o_s}/${o_s}.${top_tag}" ]; then
            d_source_dir="${INSTALL_SRC}/$o_s"
            return 0
        fi
    else                                          # Environment variable not set
        if [ -z "$d_source_dir" ]; then           # Should be set in InitVars
            d_source_dir="/cdrom"
        fi
        for check_point in "$start_dir" "$d_source_dir" "$launch_dir"; do
            if [ -f "${check_point}/${o_s}.${top_tag}" ]; then
                d_source_dir="$check_point"
                return 0
            elif [ -f "${check_point}/${o_s}/${o_s}.${top_tag}" ]; then
                d_source_dir="${check_point}/$o_s"
                return 0
            fi
        done
    fi

    Alert "$is_not_correct_dir_msg" 
    return 1
}

#------------------------------------------------------------------------------
# FindPackages - Locate and verify the packages to be installed
#------------------------------------------------------------------------------
FindPackages()
{
    source_dir="$d_source_dir"      # Use default (reserved for future use)
    e_pkg_cnt=0                     # Initialize array subscript
    c_pkg_cnt=0                     # Initialize array subscript

    # Load all of the package information into arrays
    if [ -x "$source_dir" ]; then     # Exit if cd will fail
        cd "$source_dir"
    else
        Alert "$cannot_chdir_to_msg [ $source_dir ].  $quitting_msg"
        CleanUpExit 2
    fi
    ls *.PKG > /dev/null 2>&1       # Look for .PKG files
    if [ $? -ne 0 ]; then           # No error means at least one file found
        Alert "$is_not_correct_dir_msg"
        CleanUpExit 2
    else
        #------------------------------
        # Look for Easy Only packages
        #------------------------------
        for _fnd_pkg_pkg in `ls E*.PKG 2> /dev/null`; do
            if [ -z "$_fnd_pkg_pkg" ]; then
                break                         # Leave loop
            fi
            _fnd_pkg_pkg_name="`expr \"$_fnd_pkg_pkg\" : '\(.*\)\.PKG'`"    # Get package (file) name
            if [ -f "LOCALE/$catalog_dir/${_fnd_pkg_pkg_name}.STR" ]; then
                . "${source_dir}/LOCALE/$catalog_dir/${_fnd_pkg_pkg_name}.STR" # Load strings 
            else
                continue                # Get next package
            fi

            if [ -f "${_fnd_pkg_pkg_name}.VAL" ]; then
                # Run validation script
                "${source_dir}/${_fnd_pkg_pkg_name}.VAL" '' preflight 2>/dev/null &       
                wait $!                       # Need to background it to get PID
                _fnd_pkg_result=$?
                if [ $_fnd_pkg_result -eq 1 ]; then    # Error
                    if [ -r /tmp/$!.MSG ]; then
                        Interactive && TopScreen "$hit_ret_banner_msg";
                        echo
                        cat /tmp/$!.MSG           # Display messages
                        echo
                        WaitToContinue
                        rm -f /tmp/$!.MSG         # Clear to prevent concatenation
                    else
                        Alert "$an_error_msg (preflight)"  # Generic msg
                    fi
                    echo "$find_pkgs_msg"         # Let them know what we are doing
                    continue                      # Get next package
                fi                            
            else
                continue                          # Get next package
            fi

            # Just in case VAL script changed cwd
            if [ -x "$source_dir" ]; then         # Exit if cd will fail
                cd "$source_dir"
            else
                Alert "$cannot_chdir_to_msg [ $source_dir ].  $quitting_msg"
                CleanUpExit 2
            fi

            if [ -r "${_fnd_pkg_pkg_name}.SIZ" -a -s "${_fnd_pkg_pkg_name}.SIZ" ]; then
                space_req="`cat ${_fnd_pkg_pkg_name}.SIZ`"       # Space requirements in K
            else
                continue                          # Get next package
            fi

            if [ ! -f "${_fnd_pkg_pkg_name}.INS" ]; then
                continue                          # Get next package
            fi

            # If we got this far then we have a package
            e_pkg_cnt=`expr $e_pkg_cnt + 1`       # Increment package count
            eval e_package_$e_pkg_cnt="$_fnd_pkg_pkg_name" # Name saved in array of names
            eval ${_fnd_pkg_pkg_name}_size=$space_req           # Space requirements
            eval inst_$_fnd_pkg_pkg_name=''                     # Initialize Only
        done
        #------------------------------
        # Look for Custom Only packages
        #------------------------------
        for _fnd_pkg_pkg in `ls C*.PKG 2> /dev/null`; do
            if [ -z "$_fnd_pkg_pkg" ]; then
                break                             # Leave loop
            fi
            _fnd_pkg_pkg_name="`expr \"$_fnd_pkg_pkg\" : '\(.*\)\.PKG'`"    # Get package (file) name
            if [ -f "LOCALE/$catalog_dir/${_fnd_pkg_pkg_name}.STR" ]; then
                . "${source_dir}/LOCALE/$catalog_dir/${_fnd_pkg_pkg_name}.STR" # Load strings 
            else
                continue                          # Get next package
            fi

            if [ -f "${_fnd_pkg_pkg_name}.VAL" ]; then
                # Run validation script
                "${source_dir}/${_fnd_pkg_pkg_name}.VAL" '' preflight 2>/dev/null &       
                wait $!                           # Need to background it to get PID
                _fnd_pkg_result=$?
                if [ $_fnd_pkg_result -eq 1 ]; then    # Error
                    if [ -r /tmp/$!.MSG ]; then
                        Interactive && TopScreen "$hit_ret_banner_msg";
                        echo
                        cat /tmp/$!.MSG           # Display messages
                        echo
                        WaitToContinue
                        rm -f /tmp/$!.MSG         # Clear to prevent concatenation
                    else
                        Alert "$an_error_msg (preflight)"  # Generic msg
                    fi
                    echo "$find_pkgs_msg"         # Let them know what we are doing
                    continue                      # Get next package
                fi                            
            else
                continue                          # Get next package
            fi
            # Just in case VAL script changed cwd
            if [ -x "$source_dir" ]; then     # Exit if cd will fail
                cd "$source_dir"
            else
                Alert "$cannot_chdir_to_msg [ $source_dir ].  $quitting_msg"
                CleanUpExit 2
            fi

            if [ -r "${_fnd_pkg_pkg_name}.SIZ" -a -s "${_fnd_pkg_pkg_name}.SIZ" ]; then
                space_req="`/bin/cat ${_fnd_pkg_pkg_name}.SIZ`"       # Space requirements in K
            else
                continue                      # Get next package
            fi

            if [ ! -f "${_fnd_pkg_pkg_name}.INS" ]; then
                continue                      # Get next package
            fi

            # If we got this far then we have a package
            c_pkg_cnt=`expr $c_pkg_cnt + 1`       # Increment package count
            eval c_package_$c_pkg_cnt="$_fnd_pkg_pkg_name" # Name saved in array of names
            eval ${_fnd_pkg_pkg_name}_size=$space_req           # Space requirements
            eval inst_$_fnd_pkg_pkg_name=''                     # Initialize Only
        done
        if [ $e_pkg_cnt -eq 0 -a $c_pkg_cnt -eq 0 ]; then
            Alert "$is_not_correct_dir_msg"
            CleanUpExit 2
        fi
    fi
}

#------------------------------------------------------------------------------
# AskInstPkgs - Ask user if they want to install each package in turn
#------------------------------------------------------------------------------
AskInstPkgs()
{
    #------------------
    # Ask Easy Packages
    #------------------
    _ask_pkgs_cnt=1
    while [ $_ask_pkgs_cnt -le $e_pkg_cnt ]; do
        default_destination=''                       
        eval _ask_pkgs_pkg_name="\$e_package_$_ask_pkgs_cnt"
        . "$source_dir/LOCALE/$catalog_dir/${_ask_pkgs_pkg_name}.STR"

        eval _ask_pkgs_default="\${inst_$_ask_pkgs_pkg_name:-$yes_msg}" 

        # AskYesNo returns "yes" or "no" in $reply
        AskYesNo "$install_prompt_msg" "$install_help_msg" "$_ask_pkgs_default"

        eval inst_$_ask_pkgs_pkg_name="$reply" 
        
        eval _ask_inst_do_inst="\$inst_$_ask_pkgs_pkg_name"
        if [ "$_ask_inst_do_inst" = "$yes_msg" ]; then
            SetTopDir
            eval d_${_ask_pkgs_pkg_name}_dir=\${d_${_ask_pkgs_pkg_name}_dir:-"$d_dest_dir"}

            # The AskDir function will modify the value of <_ask_pkgs_pkg_name>_dir. 
            AskDir "$destination_prompt_msg" "$destination_help_msg" "$_ask_pkgs_pkg_name" W 
            if [ $? -ne 0 ]; then                 # Bad Dir - go back
                continue                          # Ask if they want to install
            fi
            eval d_dest_dir=\$${_ask_pkgs_pkg_name}_dir   # For SetTopDir
        fi
        _ask_pkgs_cnt=`expr $_ask_pkgs_cnt + 1`
    done
    #-------------------------
    # Ask Custom Only Packages
    #-------------------------
    _ask_pkgs_cnt=1
    while [ $_ask_pkgs_cnt -le $c_pkg_cnt ]; do
        default_destination=''                       
        eval _ask_pkgs_pkg_name="\$c_package_$_ask_pkgs_cnt"
        . "$source_dir/LOCALE/$catalog_dir/${_ask_pkgs_pkg_name}.STR"

        eval _ask_pkgs_default="\${inst_$_ask_pkgs_pkg_name:-$yes_msg}" 

        # AskYesNo returns "yes" or "no" in $reply
        AskYesNo "$install_prompt_msg" "$install_help_msg" "$_ask_pkgs_default"

        eval inst_$_ask_pkgs_pkg_name="$reply" 
        
        eval _ask_inst_do_inst="\$inst_$_ask_pkgs_pkg_name"
        if [ "$_ask_inst_do_inst" = "$yes_msg" ]; then
            SetTopDir
            eval d_${_ask_pkgs_pkg_name}_dir=\${d_${_ask_pkgs_pkg_name}_dir:-"$d_dest_dir"}

            # The AskDir function will modify the value of <_ask_pkgs_pkg_name>_dir. 
            AskDir "$destination_prompt_msg" "$destination_help_msg" "$_ask_pkgs_pkg_name" W 
            if [ $? -ne 0 ]; then                 # Bad Dir - go back
                continue                          # Ask if they want to install
            fi
            eval d_dest_dir=\$${_ask_pkgs_pkg_name}_dir   # For SetTopDir
        fi
        _ask_pkgs_cnt=`expr $_ask_pkgs_cnt + 1`
    done
        
    return 0
}

#------------------------------------------------------------------------------
#   AskYesNo - keep asking a yes/no question till user answers yes/no
#   $3 is default, assumes help will be provided by all calling routines
#------------------------------------------------------------------------------
AskYesNo()
{
    if [ $# -ne 3 ]; then
        AssertErr "Incorrect argument count in AskYesNo()."
    fi

    _ask_y_n_prompt="$1"                         # Local $prompt
    _ask_y_n_help="$2"                           # Local $help
    _ask_y_n_default="$3"                        # Local $default (yes/no)
    if [ "$_ask_y_n_default" != "$yes_msg" -a "$_ask_y_n_default" != "$no_msg" ]; then
        AssertErr "Problem with arguments passed to AskYesNo()."
    fi

    while :; do
        TopScreen;
        echo "$_ask_y_n_prompt"  
        Prompt "    [$default_equals_msg$bld$_ask_y_n_default$off] >>> "
        read _ask_y_n_answer
        ToLower _ask_y_n_answer                  # Convert to Lowercase
   
        case "$_ask_y_n_answer_lwr" in
           '')                  reply=$_ask_y_n_default; return 0 ;; # Use default
           "$yes_msg"|\
           "$yes_short_msg")    reply="$yes_msg"; return ;; # Yes
           "$no_msg"|\
           "$no_short_msg")     reply="$no_msg"; return ;;  # No
           "$help_msg"|\
           "$help_short_msg")   DoHelp "$_ask_y_n_help"; continue ;; # Help (may not return)
           "$quit_msg"|\
           "$quit_short_msg")   if [ -z "$in_quit" ]; then
                                    save_ask_y_n_prompt="$_ask_y_n_prompt"
                                    save_ask_y_n_help="$_ask_y_n_help"
                                    save_ask_y_n_default="$_ask_y_n_default"
                                    AskQuit "$yes_msg"      # Quit (or return here)
                                    _ask_y_n_prompt="$save_ask_y_n_prompt"
                                    _ask_y_n_help="$save_ask_y_n_help"
                                    _ask_y_n_default="$save_ask_y_n_default"
                                else
                                    CleanUpExit 0           # Go to exit
                                fi
                                continue                    # Ask again
                                ;;
           *)   Alert "$alert_not_valid_msg" ;;
       esac
    done
}

#------------------------------------------------------------------------------
# AssertErr - Display Alert and exit
#------------------------------------------------------------------------------
AssertErr()
{
    Alert "$assert_msg ${1:-} $quitting_msg"  
    CleanUpExit 1                               
}

#------------------------------------------------------------------------------
# BadDir - Check directory accessability, (and writablity) ( return 0 if bad)
#------------------------------------------------------------------------------
BadDir()
{
    # Usage: BadDir <dir_name> <pkg_name> [<write_req>]

    if [ $# -eq 3 ]; then
        if [ $3 != W ]; then
            AssertErr "Problem with argument list in BadDir()." 
        fi
    fi

    _bad_dir_arg_cnt=$#
    _bad_dir_dir="$1"
    _bad_dir_pkg_name="$2"
    
    echo "$checking_directory_msg"                # This takes a bit of time - warn them

    # Prevent user from getting the shell to do nasty things here.
    _bad_dir_dir="`echo "$_bad_dir_dir" | cut -d';' -f1`"           # Remove any junk after a ';'
    _bad_dir_dir="`echo "$_bad_dir_dir" | cut -d'|' -f1`"           # Remove any junk after a '|'

    # Use shell to expand path.  Allows for $HOME stuff
    _bad_dir_dir="`eval echo \"$_bad_dir_dir\"`"  # Expand path
    
    if [ $? -ne 0 -o "`expr \"$_bad_dir_dir\"// : '.*~.*'`" -ne 0 ]; then
        Alert "$badly_formed_path_msg" 
        return 0                                  # Path must contain bad chars
    fi

    if [ "`expr \"$_bad_dir_dir\"// : '\/.*'`" -eq 0 ]; then
        _bad_dir_dir="$start_dir/$_bad_dir_dir"   # Prepend path if not absolute
    fi

    if [ ! -d "$_bad_dir_dir" ]; then
        if [ "$_bad_dir_arg_cnt" -eq 3 ]; then    # Write privileges required
            if MakeDir "$_bad_dir_dir"; then      # Try to create dir
                :                                 # Remember to remove this dir later
            else                                  
                Alert "$cannot_create_dir_msg [$_bad_dir_dir]." 
                return 0                          # Return TRUE (is bad dir)
            fi
        else                                      # We should not create dir
            Alert "$dir_not_exist_msg [$_bad_dir_dir]."    
            return 0                              # Return TRUE (is bad dir)
        fi
    fi                                            # Dir must exist

    chmod u+rwx "$_bad_dir_dir" 2> /dev/null      # Ensure full access to dir

    #------------------------------------------------------------------------
    # If destination directory is not accessable then display alert and return
    #------------------------------------------------------------------------
    if [ ! -x "$_bad_dir_dir" ]; then
        Alert "$cannot_chdir_to_msg [$_bad_dir_dir]"       # Cannot chdir to directory
        return 0                                  # Return TRUE (is bad dir)
    else
        cd "$_bad_dir_dir" 2> /dev/null           
        _bad_dir_dir="`/bin/pwd`"                 # Get REAL path (resolves sym-links, etc.)
    fi
    if [ ! -r "$_bad_dir_dir" ]; then
        Alert "$cannot_read_from_msg [$_bad_dir_dir]"      # Cannot read from directory
        return 0                                  # Return TRUE (is bad dir)
    fi

    # Must be fully expanded before calling SpaceCheck
    eval ${_bad_dir_pkg_name}_dir=\"$_bad_dir_dir\"                 
    eval d_${_bad_dir_pkg_name}_dir=\"$_bad_dir_dir\"  
    
    #------------------------------------------------------------------------
    # If two args then test for writability and free space
    #------------------------------------------------------------------------
    if [ "$_bad_dir_arg_cnt" -eq 3 ]; then        # Write test request
        if [ -w "$_bad_dir_dir" ]; then           # If we can write to the dir
            $mkdir_cmd "$mkdir_opt" "${_bad_dir_dir}/$$" 2> /dev/null  
            if [ -d "${_bad_dir_dir}/$$" ]; then # Double Check!!
                rmdir "${_bad_dir_dir}/$$" 2> /dev/null        # Clean up tmp dir
            else                                  # Don't be fooled by CD-ROMs
                Alert "$cannot_write_to_msg [$_bad_dir_dir]"   # Cannot write to directory
                return 0                          # Return TRUE (is bad dir) 
            fi
        else
            Alert "$cannot_write_to_msg [$_bad_dir_dir]"   # Cannot write to directory
            return 0                              # Return TRUE (is bad dir)
        fi
        #--------------------------------------------------------------------
        # Check for free space on volume
        #--------------------------------------------------------------------
        if SpaceCheck "$_bad_dir_dir" "$_bad_dir_pkg_name"; then
            if [ "${we_made_dir:-FALSE}" = TRUE ]; then    # Only if we made the dir
                cd "$start_dir"                   # In case we're in the dir to be removed
                rmdir "$_bad_dir_dir" 2> /dev/null
            fi
            return 1                              # Return FALSE (not bad dir)
        else
            if [ "${we_made_dir:-FALSE}" = TRUE ]; then    # Only if we made the dir
                cd "$start_dir"                   # In case we're in the dir to be removed
                rmdir "$_bad_dir_dir" 2> /dev/null
            fi
            return 0                              # Return TRUE (is bad dir)
        fi
    fi

    return 1                                      # Return FALSE (not bad dir)
}

#------------------------------------------------------------------------------
# Chdir - Try to chdir - if fail then display message and exit
#------------------------------------------------------------------------------
# Only good for ksh - not used in this version  - I WANTED to use ksh 
#------------------------------------------------------------------------------
Chdir()
{
    if [ $# -ne 1 ]; then
        AssertErr "Incorrect number of args passed to Chdir()."
    fi
    if [ -z "$1" ]; then
        AssertErr "NULL path passed to Chdir()."
    fi

    if cd "$1" 2> /dev/null; then
        return
    else
        Alert "$cannot_chdir_to_msg [ $1 ].  $quitting_msg" 
        CleanUpExit 2
    fi
}

#------------------------------------------------------------------------------
#   CleanUpExit -  Clean up and partially installed files then call exit
#   with the status in arg[1].
#------------------------------------------------------------------------------
CleanUpExit()
{
trap 'echo $done_msg; exit 0' 1 2 3 15            # Just exit if we get another signal
        #   exit status 0 is normal
        #   exit status 1 is ASSERTION ERROR
        #   exit status 2 is I/O ERROR, parameter ERROR
        #   exit status 3 is SIGNAL

    if [ $# -ne 1 ]; then
        AssertErr "Incorrect arg count in CleanUpExit()."
    fi
    _exit_status=$1

    echo "$off"                                   # Undo any screen effects
    /bin/stty "$save_stty"                        # Reset to original settings

    # No cleanup required if status 0 
    if [ $_exit_status -eq 0 ]; then
        exit 
    fi

    if [ "$o_s" = "HPUX" -a $_exit_status -ne 99 ]; then 
        # Set a signal handler to protect us and then kill everyone in our proc grp
        # (HP-UX 9.01 does not support killing negative PIDs; can you say BUG!))
        trap 'CleanUpExit 99' 15                  # Run this func again on SIGTERM
        kill -15 0  > /dev/null 2>&1              # Send SIGTERM to init (sent to all in proc grp)
    else
        kill -15 -$! > /dev/null 2>&1             # Kill the currently running install script
    fi

    # Otherwise clean up any installed files
    Prompt "$cleaning_up_msg"                   

    # Make sure those pesky kids are dead
    wait $! > /dev/null 2>&1

    if [ ${_do_inst_status:-0} -eq 99 ]; then
        "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" remove >/dev/null 2>&1
    fi
    # Add any new cleanup code here.     

    echo "$done_msg"               
    exit $_exit_status
}

#------------------------------------------------------------------------------
# Confirm - Return TRUE (0) to continue; FALSE (1) to make changes
#------------------------------------------------------------------------------
Confirm()
{
    # Catch case where no packages are selected
    #------------------------
    # Check for Easy Packages
    #------------------------
    e_confirm_cnt=1
    while [ $e_confirm_cnt -le $e_pkg_cnt ]; do
        eval _confirm_pkg_name="\$e_package_$e_confirm_cnt"
        eval _confirm_install="\$inst_$_confirm_pkg_name"
        if [ "$_confirm_install" = "$yes_msg" ]; then
            break                                 # There is at least one package to install
        fi
        e_confirm_cnt="`expr $e_confirm_cnt + 1`"
    done
    #------------------------
    # Check for Easy Packages
    #------------------------
    c_confirm_cnt=1
    while [ $c_confirm_cnt -le $c_pkg_cnt ]; do
        eval _confirm_pkg_name="\$c_package_$c_confirm_cnt"
        eval _confirm_install="\$inst_$_confirm_pkg_name"
        if [ "$_confirm_install" = "$yes_msg" ]; then
            break                                 # There is at least one package to install
        fi
        c_confirm_cnt="`expr $c_confirm_cnt + 1`"
    done

    if [ $e_confirm_cnt -gt $e_pkg_cnt -a $c_confirm_cnt -gt $c_pkg_cnt ]; then  
        Alert "$no_packages_selected_msg"
        return 1
    fi
        
    case "$inst_mode" in 
        "$custom_msg"|\
        "$easy_msg")    _confirm_prompt="$confirm_install_msg" ;;
        *)              AssertErr "Invalid installation mode in Confirm()." ;;
    esac

     # Easy Packages
    _confirm_cnt=1
    while [ $_confirm_cnt -le $e_pkg_cnt ]; do
        eval _confirm_pkg_name="\$e_package_$_confirm_cnt"
        eval _confirm_install="\$inst_$_confirm_pkg_name"
        if [ "$_confirm_install" = "$yes_msg" ]; then
            . "$source_dir/LOCALE/$catalog_dir/${_confirm_pkg_name}.STR"   # Load package strings
            _confirm_prompt="$_confirm_prompt
        $package_name_msg"
        fi
    _confirm_cnt="`expr $_confirm_cnt + 1`"
    done
    # Custom Only Packages
    _confirm_cnt=1
    while [ $_confirm_cnt -le $c_pkg_cnt ]; do
        eval _confirm_pkg_name="\$c_package_$_confirm_cnt"
        eval _confirm_install="\$inst_$_confirm_pkg_name"
        if [ "$_confirm_install" = "$yes_msg" ]; then
            . "$source_dir/LOCALE/$catalog_dir/${_confirm_pkg_name}.STR"   # Load package strings
            _confirm_prompt="$_confirm_prompt
        $package_name_msg"
        fi
    _confirm_cnt="`expr $_confirm_cnt + 1`"
    done

    _confirm_prompt="$_confirm_prompt
$do_you_want_to_continue_msg
"

    case "$inst_mode" in
        "$easy_msg"|\
        "$custom_msg")  _confirm_help="$confirm_inst_help_msg" ;;
        *)              AssertErr "Invalid installation mode in Confirm()." ;;
    esac

    AskYesNo "$_confirm_prompt" "$_confirm_help" "$yes_msg"
    case "$reply" in
        "$yes_msg") return 0 ;;                   # Proceed
        "$no_msg")  return 1 ;;                   # Start over
    esac
}

#------------------------------------------------------------------------------
# DoClear - Clear screen (if supported)
#------------------------------------------------------------------------------
DoClear()
{
    if [ "$can_clear" = TRUE ]; then
        Prompt "$clr_scrn"                        # Clear screen
    else
        # Incrementing loops in sh are toooo slooooww.
        echo
        echo; echo; echo; echo; echo; echo; echo; echo; 
        echo; echo; echo; echo; echo; echo; echo; echo; 
        echo; echo; echo; echo; echo; echo; echo; echo; 
    fi
}

#------------------------------------------------------------------------------
#   DoHelp - display help message and wait for a RETURN to continue
#------------------------------------------------------------------------------
DoHelp()
{
    if [ $# -ne 1 ]; then
        AssertErr "Incorrect arg count in DoHelp()."
    fi
    _do_help_msg="$1"
	
    DoClear;
    echo "$rev$help_side_msg$help_title_msg$help_side_msg$off"
    echo "$_do_help_msg" 
    echo
    echo 
    echo "$help_trailer1_msg"
    echo "$help_trailer2_msg"
    echo

    WaitToContinue;
}

#------------------------------------------------------------------------------
# DoInstall - Installs requested packages 
#------------------------------------------------------------------------------
DoInstall()
{
    if Interactive; then
        TopScreen "$press_to_stop_msg"
    else
        echo; echo "$press_to_stop_msg"
    fi

    # Catch case where no packages are selected - assert error 
    if [ "$e_pkg_cnt" -le 0 -a "$c_pkg_cnt" -le 0 ]; then
        AssetErr "$no_packages_selected_msg"
    fi
        
    _error_occurred=FALSE                         # Initialize global
    appledir=''                                   # set to NULL
    #----------------------
    # Install Easy Packages
    #----------------------
    _do_inst_cnt=1
    while [ $_do_inst_cnt -le $e_pkg_cnt ]; do
        eval _do_inst_pkg_name="\$e_package_$_do_inst_cnt"
        eval _do_inst_install="\$inst_$_do_inst_pkg_name"
        if [ "$_do_inst_install" = "$yes_msg" ]; then
            _do_inst_destination="`eval echo \\$\${_do_inst_pkg_name}_dir`"
            appledir="${appledir:-$_do_inst_destination}"     # Set to first directory used 
            . "$source_dir/LOCALE/$catalog_dir/${_do_inst_pkg_name}.STR" # Load package strings

            # Run the package install script
            _do_inst_status=99
            "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" install 2>/dev/null &
            wait $!
            _do_inst_status=$?

            if [ $_do_inst_status -eq 1 ]; then   # Error
                _error_occurred=TRUE              # Global Flag
                if [ -r /tmp/$!.MSG ]; then
                    Interactive && TopScreen "$hit_ret_banner_msg";
                    echo; cat /tmp/$!.MSG; echo         # Display messages
                    WaitToContinue
                    rm -f /tmp/$!.MSG             # Clear to prevent concatenation
                else
                    Alert "$an_error_msg (install)"  # A generic message
                    echo
                fi
                "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" remove 2>/dev/null &
                wait $!
                if [ $? -ne 0 ]; then
                    if [ -r /tmp/$!.MSG ]; then
                        Interactive && TopScreen "$hit_ret_banner_msg";
                        echo; cat /tmp/$!.MSG; echo
                        WaitToContinue
                        rm -f /tmp/$!.MSG
                    else
                        Alert "$an_error_msg (remove)"  # A generic message
                        echo
                    fi
                fi
                if [ $_do_inst_cnt -lt $e_pkg_cnt ]; then    # Still more packages
                    AskQuit "$yes_msg"            # Quit (or return here)
                fi
            else                                  # Package installed without error
                _do_inst_status=99
                "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" verify 2>/dev/null &
                wait $!
                _do_inst_status=$?
                if [ $_do_inst_status -eq 1 ]; then   # Verify failed
                    _error_occurred=TRUE              # Global Flag
                    if [ -r /tmp/$!.MSG ]; then
                        Interactive && TopScreen "$hit_ret_banner_msg";
                        echo; cat /tmp/$!.MSG; echo
                        WaitToContinue
                        rm -f /tmp/$!.MSG
                    else
                        Alert "$an_error_msg (verify)"  # A generic message
                        echo
                    fi
                    "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" remove 2>/dev/null &
                    wait $!
                    if [ $? -ne 0 ]; then         # Remove failed
                        if [ -r /tmp/$!.MSG ]; then
                            Interactive && TopScreen "$hit_ret_banner_msg";
                            echo; cat /tmp/$!.MSG; echo
                            WaitToContinue
                            rm -f /tmp/$!.MSG
                        else
                            Alert "$an_error_msg (remove)" # A generic message
                            echo
                        fi
                    fi
                else
                    # Put entry in AliasList file  
                    if [ -d "${appledir}/lib" -a \
                        -w "${appledir}/lib" -a \
                        "$_do_inst_destination" != "$appledir" -a \
                        "$_do_inst_destination" != "/usr/lib" ]; then # One small hack
                        echo "nonpersistant|home|$_do_inst_destination" >> "${appledir}/lib/AliasList" 2> /dev/null
                    fi
                fi  
            fi
        fi
        _do_inst_cnt=`expr $_do_inst_cnt + 1`
    done
    #-----------------------------
    # Install Custom Only Packages
    #-----------------------------
    if [ "$inst_mode" = "$custom_msg" ]; then
        _do_inst_cnt=1
        while [ $_do_inst_cnt -le $c_pkg_cnt ]; do
            eval _do_inst_pkg_name="\$c_package_$_do_inst_cnt"
            eval _do_inst_install="\$inst_$_do_inst_pkg_name"
            if [ "$_do_inst_install" = "$yes_msg" ]; then
                _do_inst_destination="`eval echo \\$\${_do_inst_pkg_name}_dir`"
                appledir="${appledir:-$_do_inst_destination}"     # Set to first directory used 
                . "$source_dir/LOCALE/$catalog_dir/${_do_inst_pkg_name}.STR" # Load package strings
    
                # Run the package install script
                _do_inst_status=99
                "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" install 2>/dev/null &
                wait $!
                _do_inst_status=$?
                if [ $_do_inst_status -eq 1 ]; then   # Error
                    _error_occurred=TRUE              # Global Flag
                    if [ -r /tmp/$!.MSG ]; then
                        Interactive && TopScreen "$hit_ret_banner_msg";
                        echo; cat /tmp/$!.MSG; echo   # Display messages
                        WaitToContinue
                        rm -f /tmp/$!.MSG             # Clear to prevent concatenation
                    else
                        Alert "$an_error_msg (install)"  # A generic message
                        echo
                    fi
                    "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" remove 2>/dev/null &
                    wait $!
                    if [ $? -ne 0 ]; then
                        if [ -r /tmp/$!.MSG ]; then
                            Interactive && TopScreen "$hit_ret_banner_msg";
                            echo; cat /tmp/$!.MSG; echo
                            WaitToContinue
                            rm -f /tmp/$!.MSG
                        else
                            Alert "$an_error_msg (remove)"  # A generic message
                            echo
                        fi
                    fi
                    if [ $_do_inst_cnt -lt $c_pkg_cnt ]; then    # Still more packages
                        AskQuit "$yes_msg"            # Quit (or return here)
                    fi
                else                                  # Package installed without error
                    _do_inst_status=99
                    "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" verify 2>/dev/null &
                    wait $!
                    _do_inst_status=$?
                    if [ $_do_inst_status -eq 1 ]; then   # Verify failed
                        _error_occurred=TRUE              # Global Flag
                        if [ -r /tmp/$!.MSG ]; then
                            Interactive && TopScreen "$hit_ret_banner_msg";
                            echo; cat /tmp/$!.MSG; echo
                            WaitToContinue
                            rm -f /tmp/$!.MSG
                        else
                            Alert "$an_error_msg (verify)"  # A generic message
                            echo
                        fi
                        "$source_dir/${_do_inst_pkg_name}.INS" "$_do_inst_destination" remove 2>/dev/null &
                        wait $!
                        if [ $? -ne 0 ]; then         # Remove failed
                            if [ -r /tmp/$!.MSG ]; then
                                Interactive && TopScreen "$hit_ret_banner_msg";
                                echo; cat /tmp/$!.MSG; echo
                                WaitToContinue
                                rm -f /tmp/$!.MSG
                            else
                                Alert "$an_error_msg (remove)" # A generic message
                                echo
                            fi
                        fi
                    else
                        # Put entry in AliasList file  
                        if [ -d "${appledir}/lib" -a \
                            -w "${appledir}/lib" -a \
                            "$_do_inst_destination" != "$appledir" -a \
                            "$_do_inst_destination" != "/usr/lib" ]; then # One small hack
                            echo "nonpersistant|home|$_do_inst_destination" >> "${appledir}/lib/AliasList" 2> /dev/null
                        fi
                    fi  
                fi
            fi
            _do_inst_cnt=`expr $_do_inst_cnt + 1`
        done
    fi
        
    if Interactive; then
        echo; WaitToContinue 
    fi

    return 0
}

#------------------------------------------------------------------------------
# Epilogue - Any last requests?
#------------------------------------------------------------------------------
Epilogue()
{
    # Should generate a report or something # DEBUG
    if [ "${_error_occurred:-FALSE}" = TRUE ]; then
        echo; echo "$installation_failed_msg"
        CleanUpExit 2
    else
        echo; echo "$install_complete_msg" 
        CleanUpExit 0 
    fi
}

#------------------------------------------------------------------------------
# GetOS - Determine system type and OS
#------------------------------------------------------------------------------
GetOS()
{
    # o_s=${O_S:-}                                  # Check environment variable
    # case $o_s in
    #     AUX|AIX|HPUX|SUN_OS|SOLARIS)  return 0 ;; # Valid OS
    # esac
    
    case "`/bin/uname -s`" in
         A/UX*)  o_s='AUX'     ;;                 # Apple
          AIX*)  o_s='AIX'     ;;                 # IBM
        SunOS*)  o_s='SUN'     ;;                 # Sun
        HP-UX*)  o_s='HPUX'    ;;                 # HP
             *)  o_s='unknown' ;;                 # unknown
    esac
    
    #--------------------------------------------------------------------------
    # Additional code to determine O_S version and release
    #--------------------------------------------------------------------------
    case $o_s in
        SUN) case "`/bin/uname -r`" in
                 4*) o_s='SUN_OS' ;;           # SunOS 
                 5*) o_s='SOLARIS'   ;;        # Solaris 2.x
                  *) o_s='unknown' ;;          # unknown
             esac
             ;;
    esac
    
    if [ "$o_s" = unknown ]; then
        Alert "[`/bin/uname -sr`]  $not_supported_os_msg." 
        CleanUpExit 2
    fi
}

#------------------------------------------------------------------------------
# GetOpts - Get command line options
#------------------------------------------------------------------------------
GetOpts()
{
    case "$argc" in                               # Use global argc
    0)  interactive=TRUE; return 0 ;;             # No options to parse
    *)  interactive=FALSE                         # noninteractive
        dest_dir="$arg1"                          # Only argument that is recognize
        inst_mode="$easy_msg"                     # Do an Easy Install
        if [ "$dest_dir" = "-help" ]; then
            echo; echo "$usage_msg"; CleanUpExit 0
        fi
        ;;
    esac
}

#------------------------------------------------------------------------------
# InitVars - O/S specific commands loaded into variables
#------------------------------------------------------------------------------
InitVars()
{
    d_inst_mode="$easy_msg" 
    d_source_dir='/cdrom'
    d_dest_dir='/usr/apple'                    

    df_cmd='/bin/df'                              
    format_cmd='/bin/fmt'                         
    mkdir_cmd='/bin/mkdir'                      
    mkdir_opt='-p'                      
    tput_cmd='/bin/tput'                      
    if `echo -n XXXX | grep n > /dev/null 2>&1`; then
        bsd_echo=FALSE
    else
        bsd_echo=TRUE
    fi

    case "$o_s" in
          AIX)    df_arg='-I'                       
                  ;;
        SUN_OS)   df_arg=''                       
                  d_dest_dir='/opt/apple'            
                  ;;
        SOLARIS)  df_arg='-k'                       
                  d_dest_dir='/opt/apple'            
                  ;;
         HPUX)    format_cmd='/bin/adjust'         
                  df_cmd='/usr/bin/bdf'             
                  df_arg=''                         
                  tput_cmd='/usr/bin/tput'          
                  ;;
          AUX)    df_arg='-B'                       
                  tput_cmd='/usr/bin/tput'          
                  format_cmd='/usr/ucb/fmt'
                  mkdir_opt=''                      
                  ;;
           *)     AssertErr "Unknown O_S type in InitVars()."
                  ;;
    esac

    # Test to determine whether we are dumb terminal or smart.
    $tput_cmd clear > /tmp/tput.test$$ 2> /dev/null
    if [ -s /tmp/tput.test$$ ]; then
        can_clear=TRUE
    else
        can_clear=FALSE
    fi
    rm -f /tmp/tput.test$$
}

#------------------------------------------------------------------------------
# Interactive - Return TRUE (0) if interactive=TRUE
#------------------------------------------------------------------------------
Interactive()
{
    if [ "$interactive" = TRUE ]; then
         return 0                                 # We ARE interactive
    else
        return 1                                  # We are NOT interactive
    fi
}

#--------------------------------------------------------------------------
# ListPackages - Let the user know what packages are available
#--------------------------------------------------------------------------
ListPackages()
{
    TopScreen "$hit_ret_banner_msg"               # Clear screen and diplay header
    echo "$here_are_the_pkgs_msg"
    #----------------------
    # Easy Only Packages
    #----------------------
    if [ "$e_pkg_cnt" -ge 1 ]; then
        echo "$here_are_the_e_pkgs_msg"
         _list_pkgs_cnt=1
        while [ $_list_pkgs_cnt -le $e_pkg_cnt ]; do
            eval _list_pkgs_pkg_name="\$e_package_$_list_pkgs_cnt"
            eval _list_pkgs_size="\$${_list_pkgs_pkg_name}_size"
            # Convert size to number of megabytes
            _list_pkgs_size="`expr $_list_pkgs_size / 1024`"  
            # Add 1 to inflate the number slightly (almost the same as rounding up)
            _list_pkgs_size="`expr $_list_pkgs_size + 1`"
    
            . "$source_dir/LOCALE/$catalog_dir/${_list_pkgs_pkg_name}.STR"   # Load package strings
            echo "        ${_list_pkgs_size}MB	$package_name_msg"
    
        _list_pkgs_cnt="`expr $_list_pkgs_cnt + 1`"
        done
    fi
    #----------------------
    # Custom Only Packages
    #----------------------
    if [ "$c_pkg_cnt" -ge 1 ]; then
        echo "$here_are_the_c_pkgs_msg"
        _list_pkgs_cnt=1
        while [ $_list_pkgs_cnt -le $c_pkg_cnt ]; do
            eval _list_pkgs_pkg_name="\$c_package_$_list_pkgs_cnt"
            eval _list_pkgs_size="\$${_list_pkgs_pkg_name}_size"
            # Use 1000 (instead of 1024) to inflate the number slightly
            _list_pkgs_size="`expr $_list_pkgs_size / 1000`"  
            if [ "$_list_pkgs_size" -eq 0 ]; then
                _list_pkgs_size='<1'
            fi
    
            . "$source_dir/LOCALE/$catalog_dir/${_list_pkgs_pkg_name}.STR"   # Load package strings
            echo "        ${_list_pkgs_size}MB	$package_name_msg"
    
        _list_pkgs_cnt="`expr $_list_pkgs_cnt + 1`"
        done
    fi
    echo
    WaitToContinue
}

#--------------------------------------------------------------------------
# LoadMountList - load (global) array with mount points and free space
#--------------------------------------------------------------------------
LoadMountList()
{
    IFS='
'
    _ld_mnt_cnt=1
    _skipped_header=''
    for _mnt_point in `$df_cmd $df_arg 2> /dev/null`; do
        if [ -z "$_skipped_header" ]; then
            _skipped_header="TRUE"
            continue                              # Skip the header
        fi
        set -- `echo X$_mnt_point | awk '{ print($4); print($6) }'`  # Prepend an X for df() bug
        if [ "${1:-NULL}" = NULL -o "${2:-NULL}" = NULL ]; then
            continue                              # Must be a bad mount - ignore
        fi
        eval free_space_$_ld_mnt_cnt="$1"
        if [ -d "$2" -a -x "$2" ]; then           # If dir exists and is searchable
            cd "$2"                               # We're pretty darn sure that the dir exists
            _tmp_dir="`/bin/pwd 2> /dev/null`"    # Get "real" path (resolve sym-links)
            if [ $? -ne 0 ]; then
                _tmp_dir="$2"                     # Use whatever df() gave us
            fi
        else
            _tmp_dir="$2"                         # Use whatever df() gave us
        fi
        eval mnt_point_$_ld_mnt_cnt="$_tmp_dir"

        _ld_mnt_cnt=`expr $_ld_mnt_cnt + 1`
    done

    MAX_MNTS=`expr $_ld_mnt_cnt - 1`              # Global
    IFS="$S_IFS"
}

#------------------------------------------------------------------------------
# MakeDir - Make directory(s) from supplied path - always absolute path
#------------------------------------------------------------------------------
MakeDir()
{
    if [ $# -ne 1 ]; then
        AssertErr "Incorrect arg count in MakeDir()."
    fi
    _mk_dir_target="$1"
    we_made_dir=FALSE                             # Global Flag

    if [ -d "$_mk_dir_target" ]; then             # Already exists
        return 0              
    fi

    $mkdir_cmd "$mkdir_opt" "$_mk_dir_target" > /dev/null 2>&1
    if [ ! -d "$_mk_dir_target" ]; then           # Could not make directory
        return 1                                  # Return Error
    fi

    we_made_dir=TRUE                              # Global Flag
    return 0                                      # Return Success
}

#------------------------------------------------------------------------------
# Prompt - Display prompt - suppress CR/LF 
#------------------------------------------------------------------------------
Prompt()
{
    if [ $# -ne 1 ]; then
        AssertErr "Incorrect arg count in Prompt()."
    fi

    if [ "$bsd_echo" = TRUE ]; then
        echo -n "$1"
    else
        echo "$1\c"
    fi
}

#------------------------------------------------------------------------------
# RemoveDir - Remove directory(s) from supplied path, if empty
#------------------------------------------------------------------------------
RemoveDir()
{
    if [ $# -ne 1 ]; then
        AssertErr "Incorrect arg count in RemoveDir()."
    fi

    _rem_dir_dir="$1"
    cd "$start_dir"                               # In case we're in the dir to be removed
    
    until [ -z "$_rem_dir_dir" ]; do
        rmdir "$_rem_dir_dir" 2> /dev/null        # Remove dir if possible
        if [ $? -ne 0 ]; then                     # Could not remove dir (not an error)
            return 0
        fi
        _rem_dir_dir="`expr \"$_rem_dir_dir\" : '\(.*\)\/.*'`"      # remove rightmost dirname
    done

    return 0                                      # Return TRUE
}

#------------------------------------------------------------------------------
# MountPoint - determine the mountpoint of a directory
#------------------------------------------------------------------------------
MountPoint()
{
    if [ $# -ne 1 ]; then
        AssertErr "Incorrect arg count in MountPoint()."
    fi

    _mnt_pnt_dir_name="$1"
    _mnt_pnt_mount_point=

    until [ -n "$_mnt_pnt_mount_point" ]; do      # Loop until mountpoint is found
        IFS='
'
        for _mnt_pnt_tmp_line in `$df_cmd $df_arg 2> /dev/null`; do
            IFS="$S_IFS"
            set -- "`echo X$_mnt_pnt_tmp_line | awk '{ print($6) }'`"
            if [ "${1:-NULL}" = "$_mnt_pnt_dir_name" ]; then
                _mnt_pnt_mount_point="$_mnt_pnt_dir_name"
                continue 2                        # Continue in 'until' loop
            fi
        done

        if [ -z "$_mnt_pnt_mount_point" ]; then            # Haven't found it
            _mnt_pnt_dir_name="`expr \"$_mnt_pnt_dir_name\" : '\(.*\)\/.*'`"  # remove rightmost dir
            # The command above will turn '/foo' into '' so we put '/' back
            if [ -z "$_mnt_pnt_dir_name" ]; then
                _mnt_pnt_dir_name='/'                      # If doesn't match now, we're hosed!
            fi
        fi
    done

    _mnt_pnt_cnt=1
    while [ $_mnt_pnt_cnt -le $MAX_MNTS ]; do
        eval _mnt_pnt_mount_point_tmp="\$mnt_point_$_mnt_pnt_cnt"
        if [ "$_mnt_pnt_mount_point" = "$_mnt_pnt_mount_point_tmp" ]; then
            return $_mnt_pnt_cnt                           # Return with the mount index
        else
            _mnt_pnt_cnt=`expr $_mnt_pnt_cnt + 1` 
        fi
    done

    Alert "$could_not_find_mnt_point_msg"         # Should have returned above
    return 99                                     # Error detected
}

#------------------------------------------------------------------------------
# SetTopDir - Sets d_dest_dir
#------------------------------------------------------------------------------
SetTopDir()
{
    if [ "${d_dest_dir:-NULL}" = NULL ]; then
        AssertErr "d_dest_dir not set in SetTopDir"
        exit 1
    fi

    if [ "`basename  \"$d_dest_dir\"`" != apple ]; then # If doesn't end in 'apple'
        _tmp_dir="`expr $d_dest_dir : '\(.*apple\)'`" # Grab everything up to the 'apple'
        if [ -n "$_tmp_dir" ]; then               # If something returned from expr
            d_dest_dir="$_tmp_dir"                # Use it
        else                                      # Else expr returned NULL
            d_dest_dir="`dirname \"$d_dest_dir\"`" # Use parent of current d_dest_dir
        fi
    fi                                        

    # Get the name of prefered dir
    if [ -n "$default_destination" ]; then        # Set in <package>.STR
        d_dest_dir="`eval echo $default_destination`"
    fi

    # If target dir does not exist create it (remove it before leaving this function)
    if [ ! -d "$d_dest_dir" ]; then
        MakeDir "$d_dest_dir"                     # Try to create dir
    fi
 
    # If d_dest_dir is not accessable then default to HOME directory
    if [ ! -d "$d_dest_dir" -o ! -x "$d_dest_dir" -o ! -w "$d_dest_dir" ]; then
        dir="`basename \"$d_dest_dir\"`"          # Get the name of prefered dir
        if [ "$dir" != "apple" ]; then
            dir="apple/$dir"                      # Add "apple" if not there
        fi
        set_top_tmp="${HOME:-/tmp}"               # Use HOME, or /tmp if not set
        d_dest_dir="${set_top_tmp}/$dir"          # Make it a sub dir of HOME
    else
        if [ "${we_made_dir:-FALSE}" = TRUE ]; then
            cd "$start_dir"                       # In case we're in the dir to be removed
            rmdir "$d_dest_dir" 2> /dev/null      # Remove dir if empty
        fi
    fi
}

#------------------------------------------------------------------------------
# SpaceCheck - calculate free space - needed space; return 1 if insufficient
#------------------------------------------------------------------------------
SpaceCheck()
{
    if [ $# -ne 2 ]; then
        AssertErr "Incorrect arg count in SpaceCheck()."
    fi
    
    _spc_chk_dir="$1"                             # Target dir

    echo "$checking_space_msg"                    # This takes some time - warn them

    LoadMountList                                 # Get current avail space
    MountPoint "$_spc_chk_dir"                    # Get mount point for arg
    mnt_num=$?                                    # Entry into mount_point[]
    if [ $mnt_num -eq 99 ]; then
        return 1                                  # Error in MountPoint function
    fi

    alloc_space=0
    #-------------------------------
    # Check for Easy Packages
    #-------------------------------
    _spc_chk_cnt=1
    while [ $_spc_chk_cnt -le $e_pkg_cnt ]; do      
        eval _spc_chk_pkg_name="\$e_package_$_spc_chk_cnt"
        eval _spc_chk_inst_var="\$inst_$_spc_chk_pkg_name"   
        if [ "$_spc_chk_inst_var" != "$yes_msg" ]; then
            _spc_chk_cnt=`expr $_spc_chk_cnt + 1`
            continue                              # Not to be installed
        fi
        var_dir="`eval echo \\$\${_spc_chk_pkg_name}_dir`"
        MountPoint "$var_dir"                     # Get mount point for arg dest
        var_mnt_num=$?                            # Entry into mount_point[]
        if [ $var_mnt_num -eq 99 ]; then
            return 1                              # Error in MountPoint function
        fi
        if [ $var_mnt_num -ne $mnt_num ]; then
            _spc_chk_cnt=`expr $_spc_chk_cnt + 1`
            continue                              # Not on same mount point
        fi
        _spc_chk_pkg_size="`eval echo \\$\${_spc_chk_pkg_name}_size`"
        alloc_space="`expr $alloc_space + $_spc_chk_pkg_size`"
        _spc_chk_cnt=`expr $_spc_chk_cnt + 1`
    done
    #-------------------------------
    # Check for Custom Only Packages
    #-------------------------------
    _spc_chk_cnt=1
    while [ $_spc_chk_cnt -le $c_pkg_cnt ]; do      
        eval _spc_chk_pkg_name="\$c_package_$_spc_chk_cnt"
        eval _spc_chk_inst_var="\$inst_$_spc_chk_pkg_name"   
        if [ "$_spc_chk_inst_var" != "$yes_msg" ]; then
            _spc_chk_cnt=`expr $_spc_chk_cnt + 1`
            continue                              # Not to be installed
        fi
        var_dir="`eval echo \\$\${_spc_chk_pkg_name}_dir`"
        MountPoint "$var_dir"                     # Get mount point for arg dest
        var_mnt_num=$?                            # Entry into mount_point[]
        if [ $var_mnt_num -eq 99 ]; then
            return 1                              # Error in MountPoint function
        fi
        if [ $var_mnt_num -ne $mnt_num ]; then
            _spc_chk_cnt=`expr $_spc_chk_cnt + 1`
            continue                              # Not on same mount point
        fi
        _spc_chk_pkg_size="`eval echo \\$\${_spc_chk_pkg_name}_size`"
        alloc_space="`expr $alloc_space + $_spc_chk_pkg_size`"
        _spc_chk_cnt=`expr $_spc_chk_cnt + 1`
    done

    eval free_space="\$free_space_$mnt_num"
    if [ $free_space -lt $alloc_space ]; then
        Alert "$not_enough_free_space_msg [ $_spc_chk_dir ]" 
        return 1
    else
        return 0
    fi
}

#------------------------------------------------------------------------------
# ToLower - Convert input string to lowercase
#------------------------------------------------------------------------------
ToLower()
{
    if [ $# -ne 1 ]; then AssertErr "ToLower"; fi

    mixed="`eval echo \\$\$1`"

    if [ -z "$mixed" ]; then 
        eval ${1}_lwr=\"\"
    else
        # Convert alphabetic to lowercase and $ to _
        eval ${1}_lwr=\"`echo $mixed | tr "[A-Z]$" "[a-z]_"`\" 
    fi
}

#------------------------------------------------------------------------------
# TopScreen - clear screen - home cursor - display help banner
#------------------------------------------------------------------------------
TopScreen()
{
    if [ $# -eq 0 ]; then
        banner="$banner_msg"                      # Set default banner
    else
        banner="$1"                               # Override default with arg1
    fi

    DoClear;                                      # Clear screen - Home cursor
    echo  "$rev$banner$off"                       # Display banner in rev_video
}

#------------------------------------------------------------------------------
# WaitToContinue - Wait for user to hit RETURN key
#------------------------------------------------------------------------------
WaitToContinue()
{
    if Interactive ; then
    while :; do
        Prompt "${hit_ret_side}${bld}${hit_return}${off}"
        while :; do
            /bin/stty -echo                       # Don't display input
            read  key
            /bin/stty echo                        # Reset to norm
            case "$key" in
            '')     return 0 ;;                   # Time to move on
            "$help_msg"|\
            "$help_short_msg")   DoHelp "$no_help_msg"; return 0 ;;
            "$quit_msg"|\
            "$quit_short_msg")   if [ -z "$in_quit" ]; then
                                     AskQuit "$yes_msg"
                                     continue 2   # Display hit_return again
                                 else
                                     CleanUpExit 0
                                 fi
                                 ;;
            esac
        done
    done
    else
        return                                     # Don't do anything in batch mode
    fi
}

#------------------------------------------------------------------------------
# Welcome - Display welcome message and wait to continue
#------------------------------------------------------------------------------
Welcome()
{
    DoClear
    echo "$welcome_msg" 
    WaitToContinue
}

#------------------------------------------------------------------------------
# WherePutAll - Ask user where they want to install everything
#------------------------------------------------------------------------------
WherePutAll()
{
    _whr_put_cnt=1
    _easy_size=0
    #-------------------
    # Easy Packages Only
    #-------------------
    while [ $_whr_put_cnt -le $e_pkg_cnt ]; do
        eval _whr_put_pkg_name="\$e_package_$_whr_put_cnt"
        eval inst_$_whr_put_pkg_name="$no_msg"   # Just pretending for a moment
        _easy_size="`eval expr $_easy_size + \\$\${_whr_put_pkg_name}_size`"
        _whr_put_cnt=`expr $_whr_put_cnt + 1`
    done

    # Generate default location based on first package (#1)
    eval _whr_put_pkg_name="\$e_package_1"       # Get name for package number 1
    eval ${_whr_put_pkg_name}_size=$_easy_size   # Total of all packages
    eval inst_${_whr_put_pkg_name}="$yes_msg"    # For BadDir call
    default_destination=''                       # Should be set in *.STR
    . "$source_dir/LOCALE/$catalog_dir/${_whr_put_pkg_name}.STR"     # Load localized strings file
    SetTopDir                                    # Sets d_dest_dir
    eval d_${_whr_put_pkg_name}_dir=\${d_${_whr_put_pkg_name}_dir:-"$d_dest_dir"} # Used in AskDir

    # The AskDir function will modify the value of ${_whr_put_pkg_name}_dir. 
    AskDir "$easy_install_prompt_msg" "$easy_install_help_msg" "$_whr_put_pkg_name" W
    if [ $? -ne 0 ]; then                        # Bad Dir - go back
        _whr_put_cnt=1
        while [ $_whr_put_cnt -le $e_pkg_cnt ]; do
            eval _whr_put_pkg_name="\$e_package_$_whr_put_cnt"
            eval inst_$_whr_put_pkg_name=''      # Re-Initialize
            _whr_put_cnt=`expr $_whr_put_cnt + 1`
        done
        return 1
    fi

    _whr_put_cnt=1
    d_dest_dir="`eval echo \\$\${_whr_put_pkg_name}_dir`"                       
    while [ $_whr_put_cnt -le $e_pkg_cnt ]; do
        eval _whr_put_pkg_name="\$e_package_$_whr_put_cnt"
        . "${source_dir}/LOCALE/$catalog_dir/${_whr_put_pkg_name}.STR"    # Load strings 
        if [ -n "$default_destination" ]; then    # Set in <package>.STR
            eval ${_whr_put_pkg_name}_dir="`echo $default_destination`"
        else
            eval ${_whr_put_pkg_name}_dir="$d_dest_dir" 
        fi
        eval inst_$_whr_put_pkg_name="$yes_msg"   # We really do want to install this package
        # Maybe we should add a line to MakeDir the directory   # DEBUG
        _whr_put_cnt=`expr $_whr_put_cnt + 1`     # Increment counter
    done

    return 0
}

###############################################################################
## Main - Main program begins                                                ##
###############################################################################

# Identify OS 
GetOS                                             # Exit on error

# Initialize platform specific vars
InitVars

# Get any options passed on the command line
GetOpts                                           # Parse command line

# Run the Non-Interactive stuff right here
if Interactive; then
    state=Welcome                                 # Get things rolling in Interactive loop
else                                              # Non-interactive Easy Install 
    if FindSourceDir; then
        FindPackages
    else
        CleanUpExit 2                             # Could not find a valid source dir
    fi
    
    # Basically do the Easy Install thang here
    _cnt=1                                        # Initialize
    _easy_size=0                                  # Initialize

    while [ $_cnt -le $e_pkg_cnt ]; do
        eval _pkg_name="\$e_package_$_cnt"
        eval inst_$_pkg_name="$no_msg"            # Just pretending for a moment
        eval ${_pkg_name}_dir=''                  
        _easy_size="`eval expr $_easy_size + \\$\${_pkg_name}_size`"
        _cnt=`expr $_cnt + 1`
    done

    # Generate default location based on first package (#1)
    eval _pkg_name="\$e_package_1"                # Get name for package number 1  
    eval ${_pkg_name}_size=$_easy_size            # Total of all packages
    eval inst_${_pkg_name}="$yes_msg"             # BadDir should think we are installing 1 pkg
    eval ${_pkg_name}_dir="$dest_dir"             # dest_dir set in GetOpts

    if BadDir "$dest_dir" "$_pkg_name" W; then    # Create dir and check space.
        CleanUpExit 2
    fi
    d_dest_dir="`eval echo \\$\${_pkg_name}_dir`" # BadDir may have expanded the pathname

    _cnt=1
    while [ $_cnt -le $e_pkg_cnt ]; do
        eval _pkg_name="\$e_package_$_cnt"
        . "${source_dir}/LOCALE/$catalog_dir/${_pkg_name}.STR"  # Load strings 
        # NOTE: d_dest_dir MUST be set; it is referenced in the *.STR files.
        if [ -n "$default_destination" ]; then    # Set in <package>.STR
            eval ${_pkg_name}_dir="`echo $default_destination`"
        else
            eval ${_pkg_name}_dir="$d_dest_dir" 
        fi

        # Run validation script
        "${source_dir}/${_pkg_name}.VAL" "`eval echo \\$\${_pkg_name}_dir`" destination 2>/dev/null &       
        wait $!                                   # Need to background it to get PID
        _pkg_result=$?
        if [ $_pkg_result -eq 1 ]; then           # Error
            if [ -r /tmp/$!.MSG ]; then
                echo
                cat /tmp/$!.MSG                   # Display messages
                echo
                rm -f /tmp/$!.MSG                 # Clear to prevent concatenation
            else
                Alert "$an_error_msg (destination)"  # Generic msg
            fi
            CleanUpExit 2
        fi                            

        eval inst_$_pkg_name='$yes_msg'           # Install ALL Easy packages
        _cnt=`expr $_cnt + 1`
    done

    state=DoInstall                               # Finsish things off in Interactive loop
fi

#-------------------------------------------------------------------------------
# Interactive experience controled by this loop
#-------------------------------------------------------------------------------

while :; do
    case "$state" in 
    Welcome)       Welcome                        # Display welcome message and wait
                   last_state=$state
                   state=FindSourceDir
                   ;;
    FindSourceDir) if FindSourceDir; then         # Locate source files - modifies source_dir
                       last_state=$state
                       state=FindPackages
                   else
                       CleanUpExit 2
                   fi
                   ;;
    FindPackages)  if FindPackages; then          # Load package info into the installer
                       last_state=$state
                       state=ListPackages
                   fi
                   ;;
    ListPackages)  ListPackages                   # Display package info 
                   last_state=$state
                   state=AskInstMode
                   ;;
    AskInstMode)   if AskInstMode; then           # Modifies inst_mode
                       last_state=$state
                       case $inst_mode in
                           $easy_msg)    state=WherePutAll ;;
                           $custom_msg)  state=AskInstPkgs ;;
                           *)            AssertErr 01 ;;
                       esac
                   fi
                   ;;
    AskInstPkgs)   if AskInstPkgs; then           # Modifies inst_*
                       last_state=$state
                       state=Confirm
                   fi
                   ;;
    WherePutAll)   if WherePutAll; then           # Modifies inst_*
                       last_state=$state
                       state=Confirm
                   else
                       state=$last_state          # Go back the the last question
                   fi
                   ;;
    Confirm)       if Confirm; then               
                      state=DoInstall            # Go forward
                   else
                       state=ListPackages         # Go Back
                   fi
                   ;;
    DoInstall)     DoInstall                      # Does not return on failure 
                   state=Epilogue
                   ;;
    Epilogue)      Epilogue
                   ;;
    Quit)          AskQuit "$yes_msg"
                   state=$last_state
                   ;;
    *)             echo "Invalid State! State = '$state'." 
                   WaitToContinue                          
                   AssertErr 00
                   ;;
    esac
done

